/*
  vc +kick13m -c99 -sc -sd -O2 -o recv7 recv7.c -lamigas
  recv7 100100 > path:to/file
*/

#include <proto/exec.h>
#include <proto/dos.h>
#include <clib/alib_protos.h>
#include <dos/dos.h>
#include <devices/serial.h>

struct DosLibrary *DOSBase;

int atoi(const char *s)
{
  int n = 0;
  while (*s)
  {
    n = 10 * n + (*s++ - '0');
  }
  return n;
}

int strlen(const char *s)
{
  int n = 0;
  while (*s++)
  {
    n++;
  }
  return n;
}

LONG read(UBYTE *buffer, LONG bytes);
LONG decode(UBYTE *buffer, LONG bytes);

int main(char *args)
{
  int retval = 20;
  if ((DOSBase = (struct DosLibrary *) OpenLibrary("dos.library", 0)))
  {
    LONG bytes = atoi(args);
    LONG msgbytes = (((bytes << 3) + 6) / 7) + 3;
    UBYTE *buffer;
    if ((buffer = AllocMem(msgbytes, 0L)))
    {
      LONG inbytes = read(buffer, msgbytes);
      if (inbytes >= 3)
      {
        LONG outbytes = decode(buffer, inbytes);
        if (outbytes >= 0)
        {
          Write(Output(), buffer, outbytes);
          retval = 0;
        }
        else
        {
          Write(Output(), "decode\n", 7);
        }
      }
      else
      {
        Write(Output(), "read\n", 5);
      }
      FreeMem(buffer, msgbytes);
    }
    else
    {
      Write(Output(), "alloc\n", 6);
    }
    CloseLibrary((struct Library *) DOSBase);
  }
  return retval;
}

LONG decode(UBYTE *buffer, LONG bytes)
{
  UBYTE *msg = buffer;
  UBYTE *out = buffer;
  ULONG n = 0;
  do
  {
    if (n++ >= bytes)
    {
      Write(Output(), "start1\n", 7);
      return -1;
    }
  }
  while (*msg++ != (UBYTE)0xF0);
  if (n++ >= bytes)
  {
    return -1;
  }
  if (*msg++ == (UBYTE)0x7D)
  {
    UBYTE output = 0;
    BYTE shift = -1;
    while (n++ < bytes)
    {
      UBYTE input = *msg++;
      if (input == 0xF7)
      {
        break;
      }
      if (input & 0x80)
      {
        Write(Output(), "high\n", 5);
        return -1;
      }
      if (shift == -1)
      {
        output |= input << 1;
        shift = 6;
      }
      else
      {
        output |= input >> shift;
        *out++ = output;
        output = (input & ((1 << shift) - 1)) << (8 - shift);
        --shift;
      }
    }
    if (shift == 6)
    {
      *out++ = output;
    }
    return out - buffer;
  }
  else
  {
    Write(Output(), "start2\n", 7);
  }
  return -1;
}

LONG read(UBYTE *buffer, LONG bytes)
{
  LONG retval = -1;
  struct MsgPort *port;
  if ((port = CreatePort(NULL, 0)))
  {
    struct IOExtSer *io;
    if ((io = (struct IOExtSer *) CreateExtIO(port, sizeof(struct IOExtSer))))
    {
      if (0 == OpenDevice("serial.device", 0, (struct IORequest *) io, 0L))
      {
        struct IOTArray term = { 0xF7F7F7F7, 0xF7F7F7F7 };
        io->IOSer.io_Command = SDCMD_SETPARAMS;
        io->io_RBufLen = ((bytes + 63) >> 6) << 6;
        io->io_Baud = 31250;
        io->io_ReadLen = 8;
        io->io_StopBits = 1;
        io->io_TermArray = term;
        io->io_SerFlags = SERF_EOFMODE | SERF_XDISABLED | SERF_RAD_BOOGIE;
        if (0 == DoIO((struct IORequest *) io))
        {
          io->IOSer.io_Command = CMD_READ;
          io->IOSer.io_Length = bytes;
          io->IOSer.io_Data = buffer;
          if (0 == DoIO((struct IORequest *) io))
          {
            retval = io->IOSer.io_Actual;
          }
          else
          {
            Write(Output(), "CMD_READ\n", 9);
          }
        }
        else
        {
          Write(Output(), "SDCMD_SETPARAMS\n", 16);
        }
        CloseDevice((struct IORequest *) io);
      }
      else
      {
        Write(Output(), "OpenDevice\n", 11);
      }
      DeleteExtIO((struct IORequest *) io);
    }
    else
    {
      Write(Output(), "ExtIO\n", 6);
    }
    DeletePort(port);
  }
  else
  {
    Write(Output(), "Port\n", 5);
  }
  return retval;
}
